
%
% parser for policy.conf file:
% builds entire policy model after a slight modification of input file:
% constraint specification in policy.conf must be moved to the end of the file
%



:- import dcg/2 from dcg.
:- import append/3 from basics.
:- import str_sub/3, string_substitute/4, str_cat/3 from string.

:- auto_table.

:- debug_ctl(hide, [name/3,
		    name_list/3,
		    opt_name_list/3,
		    opt_semicolon/2,
		    c_quote/1,
		    set/3,
		    setL/3,
		    expr/3,
		    coma_list/3,
		    opt_alias_def/3,
		    opt_attr_list/3,
		    opt_var_attr_list/3,
		    scan_file/1,
		    parse_mls_rules/2,
		    parse_mls_rule/2,
		    parse_init_sid_context_rule/2,
		    parse_flask_access_vector_rule/2,
		    parse_flask_security_class_rule/2,
		    parse_flask_initial_sid_rule/2,
		    parse_attribute_rule/2,
		    parse_type_decl_rule/2,
%		    parse_te_access_vector_rule/2,
		    parse_te_assertion_rule/2,
		    parse_type_transition_rule/2,
		    parse_type_change_rule/2,
		    parse_role_decl_rule/2,
		    parse_role_allow_rule/2,
		    parse_user_conf_rule/2,
%		    parse_rules/2,
		    parse_constraints/0
		   ]).

'C'(X,Y,Z) :- word(Y,X,Z).


model_policy(MLSFile,InFile) :-
	parse_filename(InFile, _DirName, FileName, _Ext),
%	str_cat(FileName, '.P', OutFile), 
	str_cat(FileName, '_nm.P', OutFileName), 
	str_cat('model/', OutFileName, OutFile), 
	nl, write('Modeling SELinux policy specification after '),
	write(InFile), writeln(' ...'),
	cputime(B),
	tell(OutFile),
	scan_file(MLSFile),
	parse_mls_rules(0,_),
	nl,
	abolish_all_tables,
	scan_file(InFile),
	parse_rules(0,_),
	nl,
	parse_constraints,
	told,
	cputime(E),
	T is E-B,
	write('Done in '), write(T), write(' sec. '),
	write('... Output file: '), write(OutFile),
	writeln('.'), nl.



parse_rules --> parse_rule
	      | parse_rules, opt_semicolon, parse_rule.


% parse single policy rule
parse_rule --> ['`'], parse_rule, [CQ], {!, c_quote(CQ)}
	     | parse_init_sid_context_rule, !
	     | parse_flask_access_vector_rule, !
	     | parse_flask_security_class_rule, !
	     | parse_flask_initial_sid_rule, !
	     | parse_attribute_rule, !
	     | parse_type_decl_rule, !
	     | parse_te_access_vector_rule, !
	     | parse_te_assertion_rule, !
	     | parse_type_transition_rule, !
	     | parse_type_change_rule, !
	     | parse_role_decl_rule, !
	     | parse_role_allow_rule, !
	     | parse_user_conf_rule, !
	     | parse_fs_use_rule, !
	     | parse_genfs_context_rule, !
	     | parse_net_context_rule, !.


%
% FLASK rules
%

% access vector specification
parse_flask_access_vector_rule --> [common],name(C),['{'],name_list(As),['}'],
	                           {!, write(access(common,C,As)),writeln('.')}
			         | [class],name(C),['{'],name_list(As),['}'],
	                           {!, write(access(class,C,As)),writeln('.')}	
			         | [class], name(C), [inherits], name(IC),
			           opt_name_list(As),
				   {write(inherits(C,IC)), writeln('.'),
				    As = [] -> true
				             ; write(access(class,C,As)),
				               writeln('.')}.

opt_name_list(Ns) --> ['{'], name_list(Ns), ['}']
	            | [], {Ns = []}.

% security class definition
parse_flask_security_class_rule --> [class],name(C),
	                            {fmt_write("security_class(%s).\n",
					       args(C))}.

% initial sid definition
parse_flask_initial_sid_rule --> [sid], name(SID),
	                         {fmt_write("initial_sid(%s).\n",
					    args(SID))}.



%
% TE rules
%

% attribute declaration
parse_attribute_rule --> [attribute], !, name(Id), [';'],
	                 {fmt_write("attr(%s).\n",args(Id))}.

% type declaration
parse_type_decl_rule --> [type], name(Id),
	              opt_alias_def(Aliases),
		      opt_attr_list(Atts),
		      opt_var_attr_list(VAtts), [';'],
		      {append(Atts,VAtts,AllAtts),
		       write(type_decl(Id,Aliases,AllAtts)), writeln('.')}.

opt_alias_def(Aliases) --> [alias], setL(Aliases)
	                 | [], {Aliases = []}.  % empty string

opt_attr_list(Atts) --> [','], attr_list(Atts)
	              | [], {Atts = []}.

attr_list(Atts) --> name(A), {Atts = [A]}
	          | attr_list(As), [','], name(A), {append(As,[A],Atts)}.

opt_var_attr_list(VAtts) --> name_list(VAtts), !
	                   | [], {VAtts = []}.


% TE access vector
parse_te_access_vector_rule --> [AV], {av_kind(AV)},
	                        set(STs), set(TTs),
				[':'], set(Cs), set(Ps), [';'],
			        {write_access_vector_rule(AV,STs,TTs,Cs,Ps)}.


av_kind(allow).
av_kind(auditallow).
av_kind(dontaudit).


set(S) --> ['*'], {S = '_'}
	 | setL(S)
	 | ['~'], setL(S1), {S = except(S1)}.


write_access_vector_rule(AV,STs,TTs,Cs,Ps) :-
	(
	 is_list(STs)
	 -> write_each_s_type(AV,STs,TTs,Cs,Ps)
	  ; (
	     is_list(TTs)
	     -> write_each_t_type(AV,STs,TTs,Cs,Ps)
	      ; (
		 is_list(Cs)
		 -> write_each_class(AV,STs,TTs,Cs,Ps)
		  ; (
		     is_list(Ps)
		     -> write_each_perm(AV,STs,TTs,Cs,Ps)
		      ; write(access_vector(AV,STs,TTs,Cs,Ps)),
		        writeln('.')
		    )
		)
	    )
	).
	
write_each_s_type(_AV,[],_TTs,_Cs,_Ps).
write_each_s_type(AV,[ST|STs],TTs,Cs,Ps) :-
	(
	 is_list(TTs)
	 -> write_each_t_type(AV,ST,TTs,Cs,Ps)
	  ; (
	     is_list(Cs)
	     -> write_each_class(AV,ST,TTs,Cs,Ps)
	      ; (
		 is_list(Ps)
		 -> write_each_perm(AV,ST,TTs,Cs,Ps)
		  ; write(access_vector(AV,ST,TTs,Cs,Ps)),
		    writeln('.')
		)
	    )
	),
	write_each_s_type(AV,STs,TTs,Cs,Ps).
			 

write_each_t_type(_AV,_ST,[],_Cs,_Ps).
write_each_t_type(AV,ST,[TT|TTs],Cs,Ps) :-
	(
	 is_list(Cs)
	 -> write_each_class(AV,ST,TT,Cs,Ps)
	  ; (
	     is_list(Ps)
	     -> write_each_perm(AV,ST,TT,Cs,Ps)
	      ; write(access_vector(AV,ST,TT,Cs,Ps)),
	        writeln('.')
	    )
	),
	write_each_t_type(AV,ST,TTs,Cs,Ps).

write_each_class(_AV,_ST,_TT,[],_Ps).
write_each_class(AV,ST,TT,[C|Cs],Ps) :-
	(
	 is_list(Ps)
	 -> write_each_perm(AV,ST,TT,C,Ps)
	  ; write(access_vector(AV,ST,TT,C,Ps)),
	  writeln('.')
	),
	write_each_class(AV,ST,TT,Cs,Ps).

write_each_perm(_AV,_ST,_TT,_Cs,[]).
write_each_perm(AV,ST,TT,C,[P|Ps]) :-
	write(access_vector(AV,ST,TT,C,P)),
	writeln('.'),
	write_each_perm(AV,ST,TT,C,Ps).



% TE assertion
parse_te_assertion_rule --> [neverallow], set(STs), set(TTs),
	                    [':'], set(Cs), set(Ps), [';'],
			    {write(neverallow(STs,TTs,Cs,Ps)),writeln('.')}.


% type transition
parse_type_transition_rule --> [type_transition],set(STs),set(TTs),
	                       [':'], set(Cs), name(NT), [';'],
			       {write(type_trans(STs,TTs,Cs,NT)),writeln('.')}.


% type change
parse_type_change_rule --> [type_change], set(STs), set(TTs),
	                   [':'], set(Cs), name(NT), [';'],
			   {write(type_change(STs,TTs,Cs,NT)),writeln('.')}.


%
% RBAC rules
%
% role declaration
parse_role_decl_rule --> [role], name(Id), [types], setL(Types), [';'],
	                 {write_role_decl(Id,Types)}.


write_role_decl(Id,Type) :-
	atomic(Type),
	Type \= [],!,
	write(role_decl(Id,Type)),
	writeln('.').
write_role_decl(_,[]).
write_role_decl(Id,[T|Ts]) :-
	write(role_decl(Id,T)),
	writeln('.'),
	write_role_decl(Id,Ts).



% role allow
parse_role_allow_rule --> [allow], set(CurrentRs), set(NewRs), [';'],
	                  {write(role_allow(CurrentRs,NewRs)),writeln('.')}.



%
% user configuration rules
%
parse_user_conf_rule --> [user], name(N), [roles], setL(Roles), [';'],
	                 {write_user(N,Roles)}.


write_user(N,Role) :-
	atomic(Role),
	Role \= [],!,
	write(user(N,Role)),
	writeln('.').
write_user(_,[]).
write_user(N,[R|Rs]) :-
	write(user(N,R)),
	writeln('.'),
	write_user(N,Rs).



%
% SID contexts
%

%
% security contexts for initial SIDs
%
parse_init_sid_context_rule --> [sid], name(SID), s_con((U,R,T)),
	                        {fmt_write("init_sid_context(%s,(%s,%s,%s)).\n",
					   args(SID,U,R,T))}.

s_con((U,R,T)) --> name(U), [':'], name(R), [':'], name(T).


%
% labeling for inodes in particular filesystem types
%
parse_fs_use_rule --> [fs_use_xattr], name(FS), s_con((U,R,T)),
	              {fmt_write("fs_use(xattr,%s,(%s,%s,%s)).\n",
				 args(FS,U,R,T))}
		    | [fs_use_task], name(FS), s_con((U,R,T)),
	              {fmt_write("fs_use(task,%s,(%s,%s,%s)).\n",
				 args(FS,U,R,T))}
		    | [fs_use_trans], name(FS), s_con((U,R,T)),
	              {fmt_write("fs_use(trans,%s,(%s,%s,%s)).\n",
				 args(FS,U,R,T))}.



% genfs context rules
parse_genfs_context_rule --> [genfscon], name(FS), name(P), s_con((U,R,T)),
	                     {fmt_write("genfscon(%s,'%s',(%s,%s,%s)).\n",
					args(FS,P,U,R,T))}.


% net context rules
parse_net_context_rule --> [portcon], name(ID), name(Num), s_con((U,R,T)),
		           {fmt_write("portcon(%s,%s,(%s,%s,%s)).\n",
				      args(ID,Num,U,R,T))}
			 | [netifcon], name(ID), s_con((U1,R1,T1)),
		           s_con((U2,R2,T2)),
			   {fmt_write("netifcon(%s,(%s,%s,%s),(%s,%s,%s)).\n",
				      args(ID,U1,R1,T1,U2,R2,T2))}
 		         | [nodecon], name(A1), name(A2), s_con((U,R,T)),
			   {fmt_write("nodecon('%s','%s',(%s,%s,%s)).\n",
				      args(A1,A2,U,R,T))}.




%
% constraints
%
parse_constraints :-
	findall(P,word(constrain,P,_),Ps),
	parse_constraint_rules(Ps,Cs),
	write_constraints(Cs).


parse_constraint_rules([],[]).
parse_constraint_rules([P|Ps], [C|Cs]) :-
	parse_constraint_rule(P,C),
	parse_constraint_rules(Ps,Cs).

parse_constraint_rule(P,C) :-
	parse_constraint_rule(C,P,_).


parse_constraint_rule(CR) --> [constrain], setL(C), setL(P), expr(E), [';'],
	                      {CR = (C,P,E)}.


expr(E) --> ['('], expr(E1), [')'], {E = E1}
	  | [not], expr(NE), {str_cat('not ', NE, E1), E = [E1]}
	  | [r1, Op], {role_operator(Op)}, [r2], {cat_three('R1', Op, 'R2', E1),
						  E = [E1]}
	  | [u1, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('U1', Op, CNs, E1),
						 E = [E1]}
	  | [u2, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('U2', Op, CNs, E1),
						 E = [E1]}
	  | [r1, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('R1', Op, CNs, E1),
						 E = [E1]}
	  | [r2, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 cat_three('R2', Op, CNs, E1),
						 E = [E1]}
	  | [t1, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 t_operator(Op, TOp),
						 E1 =.. [TOp, 'T1', CNs],
						 E = [E1]}
	  | [t2, Op], {operator(Op)}, setS(Ns), {upcase_var(Ns,CNs),
						 t_operator(Op, TOp),
						 E1 =.. [TOp, 'T2', CNs],
						 E = [E1]}
	  | expr(E1), [BC], {b_conn(BC)}, expr(E2), {append(E1,[BC|E2], E)}.



write_constraints(Cs) :-
	writeln(':- import member/2 from basics.'),nl, 
	writeln('constraint(C,P, T1,R1,U1, T2,R2,U2) :- '),
	write_constraint_rules(Cs).


write_constraint_rules([]) :-
	nl,nl.
write_constraint_rules([(Class,Perm,Expr)|Cs]) :-
	tab(6), write('(equiv_class(C,'), write(Class), writeln('),'),
	tab(7), write('equiv_perm(C,P,'), write(Perm), writeln(')'),
	tab(7), write('-> ('), write_expr(Expr), writeln(')'),
	tab(8), write('; true'),
	abolish_table_pred(write_expr/1),
	(Cs \= [] -> writeln('),') ; writeln(').')),
	write_constraint_rules(Cs).


write_expr([]).
write_expr([and|Es]) :-
	!, write(', '),
	write_expr(Es).
write_expr([or|Es]) :-
	!, write(' ; '),
	write_expr(Es).
write_expr([E|Es]) :-
	write(E),
	write_expr(Es).





%
% permission direction definition rules
%
parse_mls_rules --> parse_mls_rule
	          | parse_mls_rules, parse_mls_rule.



parse_mls_rule --> [sensitivity, N], {s_cat(N)}, opt_alias_list(As), [';'],
	           {write_mls_rule(sensitivity(N,As))}
		 | [dominance], setL(Order),
		   {write_mls_rule(dominance(Order))}
		 | [category], name(N), opt_alias_list(As), [';'],
	           {write_mls_rule(category(N,As))}
		 | [level], name(N), opt_cat_list(Cs), [';'],
		   {write_mls_rule(level(N,Cs))}
		 | type_pref(TypePref), name(TypeName), pair_set(Ps),
		   {write_mls_rule(perm_map(TypePref, TypeName, Ps))}.



s_cat(unclassified).
s_cat(confidential).
s_cat(secret).
s_cat(top_secret).

	      
opt_alias_list(AliasList) --> [alias], setL(AliasList)
	               | [], {AliasList = []}.


opt_cat_list(CatList) --> [':'], coma_list(CatList)
	                | [], {CatList = []}.

	      
type_pref(TP) --> [common], {TP = common}
	        | [class], {TP = class}.


perm_map(PS) --> name(ClassPerm), [':'], setL(BasePerm),
	         {PS = [(ClassPerm,BasePerm)]}
	       | perm_map(Ps), name(ClassPerm), [':'], name(BasePerm),
		 {append(Ps, [(ClassPerm,BasePerm)], PS)}.



write_mls_rule(perm_map(Pref,Class,Ps)) :- !,
	write_mls_perm_map(Pref,Class,Ps).
write_mls_rule(R) :-
	write(R), writeln('.').


write_mls_perm_map(_,_, []).
write_mls_perm_map(Pref, Class, [(_P,BP)|Ps]) :-
	BP = none,!,
	write_mls_perm_map(Pref,Class,Ps).
write_mls_perm_map(Pref, Class, [(P,BP)|Ps]) :-
	atomic(BP),!,
	write(perm_map(Pref,Class,P,BP)), writeln('.'),
	write_mls_perm_map(Pref,Class,Ps).
write_mls_perm_map(Pref, Class, [(P,BP)|Ps]) :-
	BP = [R,W],
	write(perm_map(Pref,Class,P,R)), writeln('.'),
	write(perm_map(Pref,Class,P,W)), writeln('.'),
	write_mls_perm_map(Pref,Class,Ps).


%
% Auxiliaries
%

opt_semicolon --> [';'], ! | [].


setL(S) --> name(Id), {S = Id}
	 | ['{'], name_list(S), ['}'].

/*
setL(S) --> name(Id), {S = Id}
	 | ['{'], name_list(Ns1), opt_cond_suffix(Suf), name_list(Ns2), ['}'],
	   {append(Ns1, Suf, S1), append(S1, Ns2, S)}.


opt_cond_suffix(S) --> [ifdef, '('], name(D), [','], setL(Xs), [')'],
	               {program_domain(D) -> S = Xs; S = []}
		     | [], {S = []}.
*/



setS(N) --> name(N)
	  | ['{'], name_string(N), ['}'].


name_string(NS) --> name(NS)
 	          | name_string(Ns), [N], {str_cat(Ns, N, NS)}.


name(N) --> [N], {is_name(N)}
	  | ['`'], name(N), [CQ], {c_quote(CQ)}.


name_list(Ns) --> name(Id), {Ns = [Id]}
	        | name_list(NL), name(Id), {append(NL, [Id], Ns)}
		| [], {Ns = []}
		| name_list(Ns1), setL(S), name_list(Ns2),
		  {append(Ns1,S,NL), append(NL,Ns2,Ns)}.


coma_list(XS) --> name(X), {XS = [X]}
	        | coma_list(Xs), [','], name(X),
		  {append(Xs,[X],XS)}.


pair_set(S) --> ['{'], pairs(S), ['}'], !
	      | [], {S = []}.


pairs(PS) --> pair(PS)
	    | pairs(Ps), pair(P), {append(Ps,P,PS)}.

	    
pair([(C,BP)]) --> name(C), [':'], setL(BP).



is_name(X) :-
	atom_codes(X,Codes),
	Codes = [C|Cs],
	(
	 Cs = []
	 -> not l_par(C),
	    not r_par(C),
	    not l_brace(C),
	    not r_brace(C),
	    not o_quote(C),
	    not c_quote(C),
	    not wildchar(C),
	    not coma(C),
	    not semicolon(C)
	  ; not role_operator(Codes)
	).


operator('==').
operator('!=').

t_operator('==', 'type_eq').
t_operator('!=', 'type_neq').

t_operator('==', T1,T2, TOp) :-
	TOp =.. type_eq(T1,T2).
t_operator('!=', T1,T2, TOp) :-
	TOp =.. type_neq(T1,T2).

role_operator('==').
role_operator('!=').
role_operator('eq').
role_operator('dom').
role_operator('domby').
role_operator('incomp').

b_conn('and').
b_conn('or').

o_quote(96).

c_quote(X) :-
	number(X), !,
	X = 39.
c_quote(X) :-
	atom_codes(X, [39]).


upcase_vars([],[]).
upcase_vars([X|Xs],[CX|CXs]) :-
	upcase_var(X,CX),
	upcase_vars(Xs,CXs).


upcase_var(X,CX) :-
	( X = 't1' ; X = 't2' ;
	  X = 'r1' ; X = 'r2' ;
	  X = 'u1' ; X = 'u2' )
	  -> !, atom_codes(X, [L|N]),
	     CL is L-32,
	     atom_codes(CX, [CL|N])
	   ; CX = X.



cat_three(S1,S2,S3,S) :-
	str_cat(S1, ' ', S1B),
	str_cat(S1B, S2, S12),
	str_cat(S12, ' ', S2B),
	str_cat(S2B, S3, S).
